@@ -5,9 +5,9 @@ import __init__ as test
 
 network = False
 proteome = False
-proteinParser = False
+proteinParser = True
 uniProtAPI = False
-proteinObject = True
+proteinObject = False
 isoformAlgo = False
 ALL = True
 
